@@ -22,7 +22,7 @@ typedef struct {
     int         size;
 } dir_list;
 
-typedef int sortfn (const void *, const void*);
+typedef int sortfn (const void *, const void *);
 int do_load_dir (dir_list *list, sortfn *sort, int reverse, char *filter);
 void do_sort (dir_list *list, sortfn *sort, int top, int reverse);
 dir_list *do_collect_stat (dir_list *dir, int top);
@@ -30,15 +30,34 @@ int do_reload_dir (dir_list *list, sortfn *sort, int count, int reverse, char *f
 void clean_dir (dir_list *list, int count);
 int set_zero_dir (dir_list *list);
 
+#ifdef DIR_H_INCLUDE_HANDLE_DIRENT
+int handle_dirent (dir_list *list, char *filter, struct dirent *dp,
+		   struct stat *buf1, int next_free, int *link_to_dir);
+#endif
+
 /* Sorting functions */
-int unsorted (const void *a, const void *b);
+int unsorted (const file_entry *a, const file_entry *b);
 int sort_name (const file_entry *a, const file_entry *b);
 int sort_ext (const file_entry *a, const file_entry *b);
 int sort_time (const file_entry *a, const file_entry *b);
 int sort_size (const file_entry *a, const file_entry *b);
+int sort_inode (const file_entry *a, const file_entry *b);
+
+#define SORT_TYPES 6
+typedef struct {
+    char    *sort_name;
+    int     (*sort_fn)(const file_entry *, const file_entry *);
+} sort_orders_t;
+
+sort_orders_t sort_orders [SORT_TYPES];
 
 int link_isdir (file_entry *);
+int if_link_is_exe (file_entry *file);
+
 extern int show_backups;
 extern int show_dot_files;
 extern int show_backups;
 extern int mix_all_files;
+
+char   *sort_type_to_name (sortfn *);
+sortfn *sort_name_to_type (char *type);
