@@ -1,7 +1,7 @@
 #define MAX_FIND_MENU      4   /* Maximum Menu */
-#define MAX_FIND_FLINES   11   /* Size of Window Where files are strored */ 
+#define MAX_FIND_FLINES   11   /* Size of Window Where files are stored */ 
 #define MAX_FIND_TLINES    4   /* Size of Window Where Menu and text .. */
-#define MAX_FIND_COLS     48   /* Lenght of Windows */
+#define MAX_FIND_COLS     48   /* Length of Windows */
 #define FIND_DIALOG_SIZE  MAX_FIND_FLINES+MAX_FIND_TLINES
                           /* Total size of the whole dialog */
 
@@ -16,11 +16,8 @@ typedef struct  find_list{
 }find_list;
 
 typedef struct FStack{		/* The Stack will be used to store */
-    char *dir_name;             /* the directories to serach */
+    char *dir_name;             /* the directories to search */
     struct FStack *next;        /*  single-linked */ 
 }FStack;
 
-void mi_find();
-
-
-
+void do_find(void);
