@@ -3,12 +3,18 @@ extern char *home_dir;
 #define min(x,y) (x>y ? y: x)
 #define max(x,y) (x>y ? x: y)
 
-void repaint_screen ();
-void refresh_screen ();
+void refresh_screen (void *);
 
-#ifdef ultrix
+#if defined(ultrix) || defined(AUX)
 char *strdup (char *);
 #endif
 
 /* AIX compiler doesn't understand '\e' */
 #define ESC_CHAR '\033'
+
+#ifdef USE_BSD_CURSES
+    #define xgetch x_getch
+#else
+    #define xgetch getch
+#endif
+
