@@ -22,10 +22,10 @@
 
    Links in the hypertext file are specified like this: the text that
    will be highlighted should have a leading ^A, then it comes the
-   text, then a ^B indicating that highlighing is done, then the name
+   text, then a ^B indicating that highlighting is done, then the name
    of the node you want to link to and then a ^C.
 
-   The file must contain a ^D at the beggining and at the end of the
+   The file must contain a ^D at the beginning and at the end of the
    file or the program will not be able to detect the end of file.
 
 */
@@ -35,25 +35,37 @@
 #include <sys/stat.h>
 #include <malloc.h>
 #include <errno.h>
+#include "mad.h"
 #include "color.h"
 #include "dialog.h"
 #include "util.h"
 #include "input.h"
 #include "win.h"
 #include "global.h"
-#include "help.h"
 #include "mouse.h"
+#include "key.h"	/* For mi_getch() */
+#include "help.h"
+
+
+#ifdef USE_BSD_CURSES
+    #define ACS_MAP(x) '*'
+#else
+    #define ACS_MAP(x) acs_map [x]
+#endif
 
 #define MAXLINKNAME 80
 #define HISTORY_SIZE 20
+#define HELP_WINDOW_WIDTH 60
 
-static char rcsid [] = "$Id: help.c,v 1.11 1994/10/13 04:27:35 miguel Exp $";
+static char rcsid [] = "$Id: help.c,v 1.14 1995/01/27 02:35:04 miguel Exp $";
 static char *data;		/* Pointer to the loaded data file */
 static int help_lines = 18;	/* Lines in help viewer */
 WINDOW *whelp;			/* Help Window */
 static int  history_ptr;	/* For the history queue */
 static char *main;		/* The main node */
 static char *last_shown = 0;	/* Last byte shown in a screen */
+static int end_of_node = 0;	/* Flag: the last character of the node shown? */
+static int quit;		/* Flag: quit? */
 static char *current, *start, *selected_item;
 
 static struct {
@@ -61,11 +73,13 @@ static struct {
     char *link;			/* Pointer to the selected link */
 } history [HISTORY_SIZE];
 
+/* Link areas for the mouse */
 typedef struct Link_Area {
     int x1, y1, x2, y2;
     char *link_name;
     struct Link_Area *next;
 } Link_Area;
+
 static Link_Area *link_area = NULL;
 static int inside_link_area = 0;
 
@@ -138,10 +152,10 @@ static char *move_forward2 (char *current, int lines)
     int  line;
 
     for (line = 0, p = current; *p && *p != CHAR_NODE_END; p++){
-	if (*p == '\n')
-	    line++;
 	if (line == lines)
 	    return p;
+	if (*p == '\n')
+	    line++;
     }
     return current;
 }
@@ -157,9 +171,9 @@ static char *move_backward2 (char *current, int lines)
 	    /* We reached the beginning of the node */
 	    /* Skip the node headers */
 	    while (*p != ']') p++;
-	    return p + 1;
+	    return p + 2;
 	}
-	if (*p == '\n')
+	if (*(p - 1) == '\n')
 	    line++;
 	if (line == lines)
 	    return p;
@@ -169,19 +183,32 @@ static char *move_backward2 (char *current, int lines)
 
 static void move_forward (int i)
 {
-    if (i == 1) i++;
+    if (end_of_node)
+	return;
     current = move_forward2 (current, i);
 }
 
 static void move_backward (int i)
 {
-    if (i == 1) i++;
-    current = move_backward2 (current, i);
+    current = move_backward2 (current, ++i);
 }
 
-static void null_fn (int dummy)
+static void move_to_top (int dummy)
 {
-    /* Nothing */
+    while (current > data && *current != CHAR_NODE_END)
+	current--;
+    while (*current != ']')
+	current++;
+    current = current + 1;
+    selected_item = NULL;
+}
+
+static void move_to_bottom (int dummy)
+{
+    while (*current && *current != CHAR_NODE_END)
+	current++;
+    current--;
+    move_backward (help_lines - 1);
 }
 
 static char *follow_link (char *start, char *selected_item)
@@ -205,7 +232,7 @@ static char *follow_link (char *start, char *selected_item)
 	if (p)
 	    return p;
     }
-    return "Help file format error\n\x4";	/*  */
+    return " Help file format error\n\x4";	/*  */
 }
 
 static char *select_next_link (char *start, char *current_link)
@@ -214,17 +241,14 @@ static char *select_next_link (char *start, char *current_link)
 
     if (!current_link)
 	return 0;
-    
+
     p = search_string_node (current_link, STRING_LINK_END);
     if (!p)
-	p = start;
+	return NULL;
     p = search_string_node (p, STRING_LINK_START);
-    if (p)
-	return p - 1;
-    p = search_string_node (start, STRING_LINK_START);
-    if (p)
-	return p - 1;
-    return 0;
+    if (!p)
+	return NULL;
+    return p - 1;
 }
 
 static char *select_prev_link (char *start, char *current_link)
@@ -275,7 +299,7 @@ static void end_link_area (int x, int y)
     }
 }
 
-static void clear_link_areas ()
+static void clear_link_areas (void)
 {
     Link_Area *current;
 
@@ -290,17 +314,21 @@ static void clear_link_areas ()
 static void show (char *paint_start)
 {
     char *p;
-    int  col, line;
+    int  col, line, c;
     int  painting = 1;
+    int acs = 0;	/* Flag: Alternate character set active? */
     
     line = col = 0;
-    werase (whelp);
-    wclr (whelp);
+    wclrn (whelp, 2);
+    
     clear_link_areas ();
     if (selected_item < paint_start)
 	selected_item = NULL;
+    
     for (p = paint_start; *p != CHAR_NODE_END && line < help_lines; p++){
-	if (*p == CHAR_LINK_START){
+	c = *p;
+	switch (c){
+	case CHAR_LINK_START:
 	    if (selected_item == NULL)
 		selected_item = p;
 	    if (p == selected_item)
@@ -308,23 +336,52 @@ static void show (char *paint_start)
 	    else
 		wattrset (whelp, INPUT_COLOR);
 	    start_link_area (col, line, p);
-	} else if (*p == CHAR_LINK_POINTER){
+	    break;
+	case CHAR_LINK_POINTER:
 	    painting = 0;
 	    end_link_area (col - 1, line);
-	} else if (*p == CHAR_LINK_END){
+	    break;
+	case CHAR_LINK_END:
 	    painting = 1;
 	    wattrset (whelp, REVERSE_COLOR);
-	} else if (*p == '\n'){
+	    break;
+	case CHAR_ALTERNATE:
+	    acs = 1;
+	    break;
+	case CHAR_NORMAL:
+	    acs = 0;
+	    break;
+	case CHAR_VERSION:
+	    mvwaddstr (whelp, line+2, col+2, VERSION);
+	    col += strlen (VERSION);
+	    break;
+	case '\n':
 	    line++;
 	    col = 0;
-	} else {
+	    break;
+	case '\t':
+	    col = (col/8 + 1) * 8;
+	    break;
+	default:
 	    if (!painting)
 		continue;
-	    mvwaddch (whelp, line, col, *p);
+	    if (col > HELP_WINDOW_WIDTH-1)
+		continue;
+	    
+	    wmove (whelp, line+2, col+2);
+	    if (acs){
+		if (c == ' ' || c == '.')
+		    waddch (whelp, c);
+		else
+		    waddch (whelp, ACS_MAP(c));
+	    } else
+		waddch (whelp, c);
 	    col++;
+	    break;
 	}
     }
     last_shown = p;
+    end_of_node = line < help_lines;
     wattrset (whelp, REVERSE_COLOR);
     if (selected_item >= last_shown){
 	if (link_area != NULL){
@@ -345,6 +402,17 @@ static int help_event (Gpm_Event *event)
     if (! (event->type & GPM_UP))
 	return 0;
 
+    event->y--;
+    if (event->buttons & GPM_B_RIGHT){
+	int hp;
+
+	hp = --history_ptr % HISTORY_SIZE;
+	current = start = history [hp].page;
+	selected_item = history [hp].link;
+	show (current);
+	return 0;
+    }
+
     /* Test whether the mouse click is inside one of the link areas */
     current_area = link_area;
     while (current_area)
@@ -375,9 +443,12 @@ static int help_event (Gpm_Event *event)
 	history [hp].link = current_area->link_name;
 	current = start = follow_link (current, current_area->link_name);
 	selected_item = NULL;
-    }
-    else{
-	if (event->y < help_lines/2)
+    } else{
+	if (event->y < 0)
+	    move_backward (help_lines - 1);
+	else if (event->y >= help_lines)
+	    move_forward (help_lines - 1);
+	else if (event->y < help_lines/2)
 	    move_backward (1);
 	else
 	    move_forward (1);
@@ -389,12 +460,63 @@ static int help_event (Gpm_Event *event)
     return 0;
 }
 
+static int help_cmd (void)
+{
+    /* show help */
+    {
+	int hp;
+
+	hp = history_ptr++ % HISTORY_SIZE;
+	history [hp].page = current;
+	history [hp].link = selected_item;
+	current = start = search_string (data, "[Help]") + 1;
+    }
+    selected_item = NULL;
+    return 1;
+}
+
+static int index_cmd (void)
+{
+    char *new_item;
+
+    {
+	int hp;
+
+	hp = history_ptr++ % HISTORY_SIZE;
+	history [hp].page = current;
+	history [hp].link = selected_item;
+	current = start = search_string (data, "[Help]") + 1;
+    }
+    if (!(new_item = search_string (data, "[Contents]")))
+	message (1, " Error ", " Can't find node [Contents] in help file ");
+    else
+	current = start = new_item + 1;
+    selected_item = NULL;
+    return 1;
+}
+
+static int quit_cmd (void)
+{
+    return quit = 1;
+}
+
+static int prev_node_cmd (void)
+{
+    int hp;
+
+    hp = --history_ptr % HISTORY_SIZE;
+    current = start = history [hp].page;
+    selected_item = history [hp].link;
+    return 1;
+}
+
 void interactive_display (char *filename, char *node)
 {
+    WINDOW *fkeys;
     int c;
-    int quit = 0;
     char *new_item;
-    
+
+    quit = 0;
     if ((data = load_file (filename)) == 0){
 	message (1, " Error ", " Can't open file %s \n %s ",
 		 filename, unix_error_string (errno));
@@ -404,41 +526,54 @@ void interactive_display (char *filename, char *node)
 	message (1, " Error ", " Can't find node %s in help file ", node);
 	return;
     }
-    create_dialog (60, help_lines, " Help ", "", 0);
+
+    if (help_lines > LINES - 4)
+	help_lines = LINES - 4;
+    create_dialog (HELP_WINDOW_WIDTH, help_lines, " Help ", "", 0);
     whelp = get_top_text ();
     
     selected_item = search_string_node (main, STRING_LINK_START) - 1;
-    current = start = main;
+    current = start = main + 1;
     
     for (history_ptr = 0; history_ptr < HISTORY_SIZE; history_ptr++){
 	history [history_ptr].page = current;
 	history [history_ptr].link = selected_item;
     }
 
-    push_event (3, 3, 60 + 2, help_lines + 2, (mouse_h) help_event, 0);
+    push_event (3, 2, 60 + 2, help_lines + 3, (mouse_h) help_event,
+		0, event_use_frame);
+
+    fkeys = push_fkey (1);
+    define_label_quit (fkeys, 1, "Help", help_cmd);
+    define_label_quit (fkeys, 2, "Index", index_cmd);
+    define_label_quit (fkeys, 3, "Prev", prev_node_cmd);
+    define_label (fkeys, 4, "", 0);
+    define_label (fkeys, 5, "", 0);
+    define_label (fkeys, 6, "", 0);
+    define_label (fkeys, 7, "", 0);
+    define_label (fkeys, 8, "", 0);
+    define_label (fkeys, 9, "", 0);
+    define_label_quit (fkeys, 10, "Quit", quit_cmd);
 
     do {
 	show (current);
 	c = mi_getch ();
-	if (check_movement_keys (c, 1, help_lines, move_backward, move_forward,
-				 null_fn, null_fn))
+	if (check_fkeys (c))
+	    /* Nothing */;
+	else if (c != KEY_UP && c != KEY_DOWN &&
+	    check_movement_keys (c, 1, help_lines, move_backward, move_forward,
+				  move_to_top, move_to_bottom))
 	    /* Nothing */;
 	else switch (c){
-				/* goto previous node */
+
 	case 'l':
 	case KEY_LEFT:
-	    {
-	    int hp;
-
-	    hp = --history_ptr % HISTORY_SIZE;
-	    current = start = history [hp].page;
-	    selected_item = history [hp].link;
-	    }
+	    prev_node_cmd ();
 	    break;
 	    
-				/* follow link */
 	case '\n':
 	case KEY_RIGHT:
+	    /* follow link */
 	    if (!selected_item){
 		/* If there are no links, go backward in history */
 		int hp = --history_ptr % HISTORY_SIZE;
@@ -451,52 +586,84 @@ void interactive_display (char *filename, char *node)
 		hp = history_ptr++ % HISTORY_SIZE;
 		history [hp].page = current;
 		history [hp].link = selected_item;
-		current = start = follow_link (current, selected_item);
+		current = start = follow_link (current, selected_item) + 1;
 	    }
 	    selected_item = NULL;
 	    break;
 
-				/* show help */
-	case KEY_F(1):
-	    {
-		int hp;
-
-		hp = history_ptr++ % HISTORY_SIZE;
-		history [hp].page = current;
-		history [hp].link = selected_item;
-		current = start = search_string (data, "[Help]");
-	    }
-	    selected_item = NULL;
-	    break;
-
-				/* select next link */
+	case KEY_DOWN:
 	case '\t':
+	    /* select next link */
 	    new_item = select_next_link (start, selected_item);
 	    if (new_item){
 		selected_item = new_item;
 		if (selected_item >= last_shown){
-		    selected_item = NULL;
+		    if (c == KEY_DOWN)
+			move_forward (1);
+		    else
+			selected_item = NULL;
+		}
+	    } else if (c == KEY_DOWN)
+		move_forward (1);
+	    else
+		selected_item = NULL;
+	    break;
+
+	case KEY_UP:
+	case ALT ('\t'):
+	    /* select previous link */
+	    new_item = select_prev_link (start, selected_item);
+	    selected_item = new_item;
+	    if (selected_item < current || selected_item >= last_shown){
+		if (c == KEY_UP)
+		    move_backward (1);
+		else{
+		    if (link_area != NULL)
+			selected_item = link_area->link_name;
+		    else
+			selected_item = NULL;
 		}
 	    }
 	    break;
 
-				/* select previous link */
-	case 'b':
-	    new_item = select_prev_link (start, selected_item);
-	    selected_item = new_item;
-	    if (selected_item < current || selected_item >= last_shown){
-		if (link_area != NULL)
-		    selected_item = link_area->link_name;
-		else
-		    selected_item = NULL;
+	case 'n':
+	    /* Next node */
+	    new_item = current;
+	    while (*new_item && *new_item != CHAR_NODE_END)
+		new_item++;
+	    if (*++new_item == '['){
+		while (*new_item != ']')
+		    new_item++;
+		current = new_item + 2;
+		selected_item = NULL;
 	    }
 	    break;
-				/* quit */
-	case KEY_F(10):
+
+	case 'p':
+	    /* Previous node */
+	    new_item = current;
+	    while (new_item > data + 1 && *new_item != CHAR_NODE_END)
+		new_item--;
+	    new_item--;
+	    while (new_item > data && *new_item != CHAR_NODE_END)
+		new_item--;
+	    while (*new_item != ']')
+		new_item++;
+	    current = new_item + 2;
+	    selected_item = NULL;
+	    break;
+
+	case 'c':
+	    index_cmd ();
+	    break;
+
 	case ESC_CHAR:
+	    /* quit */
 	    quit = 1;
 	}
     } while (!quit);
+    clear_link_areas ();
+    pop_fkey (fkeys);
     pop_event ();
     free (data);
     destroy_dialog ();
