@@ -12,13 +12,14 @@ typedef struct {
     int  mark;			/* The mark position */
     int  first_shown;		/* Index of the first shown character */
     int  current_max_len;	/* Maximum length of input line */
-    int  field_len;		/* Lenght of the editing field */
+    int  field_len;		/* Length of the editing field */
     int  color;			/* color used */
     int  first;			/* Is first keystroke? */
+    int  disable_update;	/* Do we want to skip updates? */
     char *buffer;		/* pointer to editing buffer */
     char *kill_save;		/* Pointer to killed data */
     Hist *history;		/* The history */
-    int  need_push;		/* need to psh the current Input on history? */
+    int  need_push;		/* need to push the current Input on history? */
     WINDOW *window;		/* pointer to window descriptor */
 } Input;
 
@@ -29,10 +30,15 @@ void update_input (Input *in);
 void new_input (Input *in);
 void push_history (Input *in, char *text);
 void stuff (Input *in, char *text, int insert_extra_space);
+void input_disable_update (Input *in);
+void input_set_prompt (Input *in, int field_len, char *prompt);
+void input_enable_update (Input *in);
+void input_set_point (Input *in, int pos);
 
-#define XCTRL(x) (x-'a'+1)
-#define ALT(x) (0x80|x)
+#define XCTRL(x) (x & 31)
+#define ALT(x) (0x200|x)
 
 /* Flags for destroy_input */
 #define IN_KEEP_BUFFER 1
 #define IN_NORMAL 0
+
