@@ -1,10 +1,14 @@
 #define MAXMENU 15
 #define MAXQUERY 10
 
-/* Must be powers of 2 (they are bit flags */
-#define BAR_ERROR 1
-#define BAR_FILL  2
-typedef void (* callfn) (int);
+/* Must be powers of 2 (they are bit flags) */
+#define BAR_ERROR      0x1
+#define BAR_FILL       0x2
+#define BAR_NO_REFRESH 0x4
+typedef void (*callfn) ();
+/* FIXME: We have to leave this type ambiguous, because `callfn' is
+   used both for functions that take an argument and ones that don't.
+   That ought to be cleared up. */
 
 typedef struct {
     char *text;
