@@ -1,5 +1,6 @@
 /* Various utilities
-   Copyright (C) 1994 Miguel de Icaza, Mauricio Plaza.
+   Copyright (C) 1994 Miguel de Icaza, Janne Kukonlehto, Mauricio Plaza.
+
    The file_date routine is mostly from GNU's fileutils package,
    written by Richard Stallman and David MacKenzie.
 
@@ -18,13 +19,10 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include <stdio.h>
 #include <stdlib.h>
-#ifdef HAVE_UNISTD_H
 #include <sys/types.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#ifdef HAVE_LIBGEN_H
-#include <libgen.h>		/* For regcmp and regex */
-#endif
 #include <signal.h>		/* my_system */
 #include <limits.h>		/* INT_MAX */
 #include <sys/time.h>		/* select: timeout */
@@ -39,25 +37,29 @@
 #include <grp.h>
 #include <string.h>
 #include <ctype.h>
+#if defined(AIX) || defined(_AIX) || defined(__aix__)
+   #include <sys/select.h>
+#endif
+
+#include "fsusage.h"
+#include "mountlist.h"
+
 /* From dialog.h (not wanting to include it as
    it requires including a lot of other files, too) */
 int message (int error, char *header, char *text, ...);
 
-#if defined(linux) || defined(gnu_regex)
-#ifdef use_local_rx
-#include "rx.h"
-#else
-#include <regex.h>
-#endif
-#endif
+#include "mad.h"
+#include "regex.h"
 #include "util.h"
 #include "global.h"
-#include "gc.h"
 #include "profile.h"
+#include "user.h"		/* expand_format */
+typedef struct Panel Panel;
+#include "tree.h"		/* tree_add_entry and tree_remove_entry for my_mkdir and my_rmdir */
 
-static char rcsid [] = "$Id: util.c,v 1.13 1994/10/13 04:28:02 miguel Exp $";
+static char rcsid [] = "$Id: util.c,v 1.16 1995/01/27 02:36:37 miguel Exp $";
 
-static char app_text [] = "Midnight-Commander";
+char app_text [] = "Midnight-Commander";
 
 int easy_patterns = 1;
 int align_extensions = 1;
@@ -72,6 +74,17 @@ char *strdup (char *s)
 }
 #endif
 
+int is_printable (int c)
+{
+    extern int eight_bit_clean;
+
+    c &= 0xff;
+    if (eight_bit_clean)
+	return ((c >31 && c < 127) || c >= 160);
+    else
+	return (c > 31 && c < 127);
+}
+
 /* Returns the message dimensions (lines and columns) */
 int msglen (char *text, int *lines)
 {
@@ -82,10 +95,11 @@ int msglen (char *text, int *lines)
 	if (*text == '\n'){
 	    line_len = 0;
 	    (*lines)++;
+	} else {
+	    line_len++;
+	    if (line_len > max)
+		max = line_len;
 	}
-	line_len++;
-	if (line_len > max)
-	    max = line_len;
     }
     return max;
 }
@@ -106,9 +120,9 @@ char *trim (char *s, char *d, int len)
 
 char *name_trunc (char *txt, int trunc_len)
 {
-    static char x [150];
+    static char x [MAXPATHLEN+MAXPATHLEN];
     int    txt_len;
-    
+
     if (trunc_len > sizeof (x)-1){
 	fprintf (stderr, "name_trunc: too big");
 	trunc_len = sizeof (x)-1;
@@ -118,9 +132,10 @@ char *name_trunc (char *txt, int trunc_len)
 	return txt;
 
     if (tilde_trunc){
-	strncpy (x, txt, trunc_len/2);
-	strncpy (x+(trunc_len/2), txt+txt_len-(trunc_len/2), trunc_len/2);
-	x [trunc_len/2] = '~';
+	int y = trunc_len % 2;
+	strncpy (x, txt, (trunc_len/2)+y);
+	strncpy (x+(trunc_len/2)+y, txt+txt_len-(trunc_len/2), trunc_len/2);
+	x [(trunc_len/2)+y] = '~';
     } else {
 	strncpy (x, txt, trunc_len-1);
 	x [trunc_len-1] = '>';
@@ -201,7 +216,7 @@ char *string_perm (mode_t mode_bits)
     if (ismode (mode_bits, S_ISGID)) mode [6] = 's';
     if (ismode (mode_bits, S_IFCHR)) mode [0] = 'c';
     if (ismode (mode_bits, S_IFBLK)) mode [0] = 'b';
-    if (ismode (mode_bits, S_ISVTX)) mode [0] = 'i';
+    if (ismode (mode_bits, S_ISVTX)) mode [9] = 't';
     if (ismode (mode_bits, S_IFLNK)) mode [0] = 'l';
     if (ismode (mode_bits, S_IFIFO)) mode [0] = 's';
     return mode;
@@ -219,23 +234,38 @@ char *strip_home(char *dir)
 	return dir;
 }
 
-#define MI_RE_COMP re_comp
-#define MI_RE_EXEC re_exec
-/* 1 if string matches
-   0 if string doesn't match
-   -1 if error in pattern */
+static char *maybe_start_group (char *d, int do_group, int *was_wildcard)
+{
+    if (!do_group)
+	return d;
+    if (*was_wildcard)
+	return d;
+    *was_wildcard = 1;
+    *d++ = '\\';
+    *d++ = '(';
+    return d;
+}
 
-int regexp_match (char *pattern, char *string, int match_type)
+static char *maybe_end_group (char *d, int do_group, int *was_wildcard)
+{
+    if (!do_group)
+	return d;
+    if (!*was_wildcard)
+	return d;
+    *was_wildcard = 0;
+    *d++ = '\\';
+    *d++ = ')';
+    return d;
+}
+
+/* If shell patterns are on converts a shell pattern to a regular
+   expression. Called by regexp_match and mask_rename. */
+char *convert_pattern (char *pattern, int match_type, int do_group)
 {
     char *s, *d;
-    char new_pattern [100];
-    char *x;
-#if defined(linux) || defined (gnu_regex)
-    static regex_t r;
-    int    rval;
-#endif
+    static char new_pattern [100];
+    int was_wildcard = 0;
 
-    x = pattern;
     if (easy_patterns){
 	d = new_pattern;
 	if (match_type == match_file)
@@ -243,53 +273,61 @@ int regexp_match (char *pattern, char *string, int match_type)
 	for (s = pattern; *s; s++, d++){
 	    switch (*s){
 	    case '*':
+		d = maybe_start_group (d, do_group, &was_wildcard);
 		*d++ = '.';
 		*d   = '*';
 		break;
 		
 	    case '?':
-		*d++ = '.';
+		d = maybe_start_group (d, do_group, &was_wildcard);
+		*d = '.';
 		break;
 		
 	    case '.':
+		d = maybe_end_group (d, do_group, &was_wildcard);
 		*d++ = '\\';
 		*d   = '.';
+		break;
+
 	    default:
+		d = maybe_end_group (d, do_group, &was_wildcard);
 		*d = *s;
+		break;
 	    }
 	}
+	d = maybe_end_group (d, do_group, &was_wildcard);
 	if (match_type == match_file)
 	    *d++ = '$';
 	*d = 0;
-	pattern = new_pattern;
+	return new_pattern;
+    } else
+	return pattern;
+}
+
+/* 1 if string matches
+   0 if string doesn't match
+   -1 if error in pattern */
+
+int regexp_match (char *pattern, char *string, int match_type)
+{
+    static regex_t r;
+    static char *old_pattern = NULL;
+    static int old_type;
+    int    rval;
+
+    if (!old_pattern || strcmp (old_pattern, pattern) || old_type != match_type){
+	if (old_pattern){
+	    regfree (&r);
+	    free (old_pattern);
+	}
+	pattern = convert_pattern (pattern, match_type, 0);
+	if (regcomp (&r, pattern, REG_EXTENDED|REG_NOSUB))
+	    return -1;
+	old_pattern = strdup (pattern);
+	old_type = match_type;
     }
-#if defined(linux) || defined(gnu_regex)
-    if (regcomp (&r, pattern, REG_EXTENDED|REG_NOSUB))
-	return -1;
     rval = !regexec (&r, string, 0, NULL, 0);
-    regfree (&r);
     return rval;
-#else
-#ifdef HAVE_RE_EXEC
-    if (re_comp (pattern)){
-	return -1;
-    }
-    return re_exec (string);
-#else /* no re_exec */
-#ifdef HAVE_LIBGEN_H
-    s = regcmp (pattern, (char *) 0);
-    if (!s){
-	return -1;
-    }
-    d = regex (s, string);
-    free (s);
-    return d != 0;
-#else
-    /* No match, system without regex or re_exec */
-    return 0;
-#endif /* HAVE_LIBGEN_H */
-#endif /* HAVE_RE_EXEC */
-#endif /* defined(linux) */
 }
 
 char *extension (char *filename)
@@ -334,6 +372,7 @@ char *split_extension (char *x, int pad)
     return xbuf; */
 }
 
+#ifndef HAVE_MAD
 void *xmalloc (int size, char *source)
 {
     void *m = malloc (size);
@@ -344,6 +383,7 @@ void *xmalloc (int size, char *source)
     }
     return m;
 }
+#endif /* HAVE_MAD */
 
 int get_int (char *file, char *key, int def)
 {
@@ -368,6 +408,7 @@ char *load_file (char *filename)
     FILE *data_file;
     struct stat s;
     char *data;
+    long read_size;
     
     if (stat (filename, &s) != 0){
 	return 0;
@@ -375,11 +416,15 @@ char *load_file (char *filename)
     if ((data_file = fopen (filename, "r")) == NULL){
 	return 0;
     }
-    data = (char *) xmalloc (s.st_size+1, "help, load_file");
-    fread (data, s.st_size, 1, data_file);
-    data [s.st_size] = 0;
+    data = (char *) xmalloc (s.st_size+1, "util, load_file");
+    read_size = fread (data, 1, s.st_size, data_file);
+    data [read_size] = 0;
     fclose (data_file);
-    return data;
+
+    if (read_size > 0)
+	return data;
+    else
+	return 0;
 }
 
 #define UID_CACHE_SIZE 200
@@ -393,7 +438,7 @@ typedef struct {
 int_cache uid_cache [UID_CACHE_SIZE];
 int_cache gid_cache [GID_CACHE_SIZE];
 
-void init_uid_gid_cache ()
+void init_uid_gid_cache (void)
 {
     int i;
 
@@ -520,12 +565,22 @@ char *icase_search (char *text, char *data)
     return 0;
 }
 
+#ifndef HAVE_BASENAME
+char *basename (char *s)
+{
+    char  *where;
+    return ((where = strrchr (s, '/')))? where + 1 : s;
+}
+#endif
+
 char *get_full_name (char *dir, char *file)
 {
     char *d = malloc (strlen (dir) + strlen (file) + 2);
 
     strcpy (d, dir);
-    strcat (d, "/");
+    if (dir [0] != '/' || dir [1] != 0)
+	strcat (d, "/");
+    file = basename (file);
     strcat (d, file);
     return d;
 }
@@ -540,7 +595,7 @@ static int old_error;		/* File descriptor of old standard error */
 /* Creates a pipe to hold standard error for a later analysis. */
 /* The pipe can hold 4096 bytes. Make sure no more is written */
 /* or a deadlock might occur. */
-void open_error_pipe ()
+void open_error_pipe (void)
 {
     if (pipe (error_pipe) < 0){
 	message (0, " Warning ", " Pipe failed ");
@@ -557,7 +612,7 @@ void open_error_pipe ()
 /* Checks for messages in the error pipe
  * and displays an warning box if needed
  */
-void check_error_pipe ()
+void check_error_pipe (void)
 {
     char error[MAX_PIPE_SIZE];
     int len = 0;
@@ -603,36 +658,78 @@ void close_error_pipe (int error, char *text)
 	msg[len] = 0;
 	close (error_pipe[0]);
     }
+    if (error < 0)
+	return;		/* Just ignore error message */
     if (text == NULL){
-	if (len == 0) return;
+	if (len == 0) return;	/* Nothing to show */
+
+	/* Show message from pipe */
 	message (error, title, msg);
     } else {
+	/* Show given text and possible message from pipe */
 	message (error, title, " %s \n %s ", text, msg);
     }
 }
 
-static struct sigaction startup_handler;
+void my_putenv (char *name, char *data)
+{
+    char *full;
 
-int my_system (char *shell, char *command)
+    full = xmalloc (strlen (name) + strlen (data) + 2, "util, my_putenv");
+    sprintf (full, "%s=%s", name, data);
+    putenv (full);
+    /* WARNING: NEVER FREE THE full VARIABLE!!!!!!!!!!!!!!!!!!!!!!!! */
+    /* It is used by putenv. Freeing it will corrupt the environment */
+}
+
+#if 0
+static void my_putenv_expand (char *name, char macro_code)
+{
+    char *data;
+
+    data = expand_format (macro_code);
+    my_putenv (name, data);
+    free (data);
+}
+
+/* Puts some status information in to the environment so that
+   processes to be executed can access it. */
+static void prepare_environment (void)
+{
+    my_putenv_expand ("MC_CURRENT_DIR", 'd');
+    my_putenv_expand ("MC_OTHER_DIR", 'D');
+    my_putenv_expand ("MC_CURRENT_FILE", 'f');
+    my_putenv_expand ("MC_OTHER_FILE", 'F');
+    my_putenv_expand ("MC_CURRENT_TAGGED", 't');
+    my_putenv_expand ("MC_OTHER_TAGGED", 'T');
+    /* MC_CONTROL_FILE has been added to environment on startup */
+}
+#endif
+
+struct sigaction startup_handler;
+
+int my_system (int as_shell_command, const char *shell, const char *command)
 {
     struct sigaction ignore, save_intr, save_quit, save_stop;
-    sigset_t child_mask, save_mask;
+    sigset_t child_mask, save_mask; 
     pid_t pid;
-    int   status;
+    int status = 0;
 
     ignore.sa_handler = SIG_IGN;
     sigemptyset (&ignore.sa_mask);
     ignore.sa_flags = 0;
-    sigaction (SIGINT,  &ignore, &save_intr);
+    
+    sigaction (SIGINT, &ignore, &save_intr);    
     sigaction (SIGQUIT, &ignore, &save_quit);
 
     /* Restore the original SIGTSTP handler, we don't want ncurses' */
     /* handler messing the screen after the SIGCONT */
     sigaction (SIGTSTP, &startup_handler, &save_stop);
-    
+
     sigemptyset (&child_mask);
     sigaddset (&child_mask, SIGCHLD);
     sigprocmask (SIG_BLOCK, &child_mask, &save_mask);
+
     if ((pid = fork ()) < 0){
 	fprintf (stderr, "\n\nfork () = -1\n");
 	return -1;
@@ -642,7 +739,15 @@ int my_system (char *shell, char *command)
 	sigaction (SIGQUIT, &save_quit, NULL);
 	sigprocmask (SIG_SETMASK, &save_mask, NULL);
 
-	execl (shell, shell, "-c", command, (char *) 0);
+	#if 0
+	prepare_environment ();
+	#endif
+	
+	if (as_shell_command)
+	    execl (shell, shell, "-c", command, (char *) 0);
+	else
+	    execlp (shell, shell, command, (char *) 0);
+
 	_exit (127);		/* Exec error */
     } else {
 	while (waitpid (pid, &status, 0) < 0)
@@ -651,17 +756,18 @@ int my_system (char *shell, char *command)
 		break;
 	    }
     }
+    sigprocmask (SIG_SETMASK, &save_mask, NULL);
     sigaction (SIGINT,  &save_intr, NULL);
     sigaction (SIGQUIT, &save_quit, NULL);
     sigaction (SIGTSTP, &save_stop, NULL);
-    sigprocmask (SIG_SETMASK, &save_mask, NULL);
+
     return status;
 }
 
 /* Since ncurses uses a handler that automatically refreshes the */
 /* screen after a SIGCONT, and we don't want this behavior when */
 /* spawning a child, we save the original handler here */
-void save_stop_handler ()
+void save_stop_handler (void)
 {
     sigaction (SIGTSTP, NULL, &startup_handler);
 }
@@ -685,7 +791,8 @@ char *unix_error_string (int error_num)
     return buffer;
 }
 
-char *tilde_expand (char *directory, int alloc_type)
+/* Returns a newly allocated string */
+char *tilde_expand (char *directory)
 {
     struct passwd *passwd;
     char *p;
@@ -693,7 +800,7 @@ char *tilde_expand (char *directory, int alloc_type)
     int  len;
     
     if (*directory != '~')
-	return directory;
+	return strdup (directory);
 
     directory++;
     
@@ -718,13 +825,10 @@ char *tilde_expand (char *directory, int alloc_type)
 
     /* If we can't figure the user name, return home_dir */
     if (!passwd)
-	return home_dir;
+	return strdup (home_dir);
 
     len = strlen (passwd->pw_dir) + strlen (p) + 2;
-    if (alloc_type == use_gc)
-	directory = gc_alloc (len);
-    else
-	directory = xmalloc (len, "tilde_expand");
+    directory = xmalloc (len, "tilde_expand");
     strcpy (directory, passwd->pw_dir);
     strcat (directory, "/");
     strcat (directory, p);
@@ -748,7 +852,7 @@ char *copy_strings (char *first,...)
 
     len++;
 
-    result = gc_alloc (len);
+    result = xmalloc (len, "copy_strings");
     va_end (ap);
     va_start (ap, first);
     strcpy (result, first);
@@ -759,3 +863,200 @@ char *copy_strings (char *first,...)
     return result;
 }
 	
+/* A function to check if we're idle.
+   Currently checks only for key presses.
+   We could also check the mouse. */
+int is_idle (void)
+{
+    /* Check for incoming key presses     *
+     * If there are any we say we're busy */
+
+    struct fd_set select_set;
+    struct timeval timeout;
+    FD_ZERO (&select_set);
+    FD_SET (0, &select_set);
+    timeout.tv_sec = 0;
+    timeout.tv_usec = 0;
+    select (FD_SETSIZE, &select_set, 0, 0, &timeout);
+    return ! FD_ISSET (0, &select_set);
+}
+
+long blocks2kilos (int blocks, int bsize)
+{
+    if (bsize > 1024){
+	return blocks * (bsize / 1024);
+    } else if (bsize < 1024){
+	return blocks / (1024 /bsize);
+    } else
+	return blocks;
+}
+
+static struct mount_entry *mount_list;
+
+void init_my_statfs (void)
+{
+#ifndef NO_INFOMOUNT
+    mount_list = read_filesystem_list (1, 1);
+#endif
+}
+
+void my_statfs (struct my_statfs *myfs_stats, char *path)
+{
+    int i, len = 0;
+
+#ifndef NO_INFOMOUNT
+    struct mount_entry *entry = NULL;
+    struct mount_entry *temp = mount_list;
+    struct fs_usage fs_use;
+
+    while (temp){
+	i = strlen (temp->me_mountdir);
+	if (i > len && strncmp (path, temp->me_mountdir, i) == 0){
+	    len = i;
+	    entry = temp;
+	}
+	temp = temp->me_next;
+    }
+
+    if (entry){
+	get_fs_usage (entry->me_mountdir, &fs_use);
+
+	myfs_stats->type = entry->me_dev;
+	myfs_stats->typename = entry->me_type;
+	myfs_stats->mpoint = entry->me_mountdir;
+	myfs_stats->device = entry->me_devname;
+	myfs_stats->avail = getuid () ? fs_use.fsu_bavail/2 : fs_use.fsu_bfree/2;
+	myfs_stats->total = fs_use.fsu_blocks/2;
+	myfs_stats->nfree = fs_use.fsu_ffree;
+	myfs_stats->nodes = fs_use.fsu_files;
+    } else
+#endif
+    {
+	myfs_stats->type = 0;
+	myfs_stats->mpoint = "unknown";
+	myfs_stats->device = "unknown";
+	myfs_stats->avail = 0;
+	myfs_stats->total = 0;
+	myfs_stats->nfree = 0;
+	myfs_stats->nodes = 0;
+    }
+}
+
+char *skip_separators (char *s)
+{
+    for (;*s; s++)
+	if (*s != ' ' && *s != '\t' && *s != ',')
+	    break;
+    return s;
+}
+
+char *skip_numbers (char *s)
+{
+    for (;*s; s++)
+	if (!isdigit (*s))
+	    break;
+    return s;
+}
+
+/* Remove all control sequences from the argument string.  We define
+ * "control sequence", in a sort of pidgin BNF, as follows:
+ *
+ * control-seq = Esc non-'['
+ *	       | Esc '[' (0 or more digits or ';' or '?') (any other char)
+ *
+ * This scheme works for all the terminals described in my termcap /
+ * terminfo databases, except the Hewlett-Packard 70092 and some Wyse
+ * terminals.  If I hear from a single person who uses such a terminal
+ * with MC, I'll be glad to add support for it.  (Dugan)
+ */
+
+char *strip_ctrl_codes (char *s)
+{
+    int i;  /* Current length of the string's correct (stripped) prefix */
+    int j;  /* Number of control characters we have skipped so far */
+
+    if (!s)
+	return 0;
+    
+    for (i = 0, j = 0; s [i+j]; ++i)
+	if (s [i+j] != ESC_CHAR){
+	    if (j)
+		s [i] = s [i+j];
+	} else {
+	    ++j;
+	    if (s [i+j++] == '[')
+		while (strchr ("0123456789;?", s [i+j++]))
+		    /* Skip the control sequence's arguments */ ;
+	    --i;
+	}
+    s[i] = 0;
+    return s;
+}
+
+#ifndef HAVE_STRCASECMP
+/* At least one version of HP/UX lacks this */
+/* Assumes ASCII encoding */
+int strcasecmp (char *s, char *d)
+{
+    register char result;
+
+    while (1){
+	if (result = (0x20 | *s) - (0x20 | *d))
+	    break;
+	if (!*s)
+	    return 0;
+	s++;
+	d++;
+    }
+    return result;
+}
+#endif /* HAVE_STRCASECMP */
+
+/* getwd is better than getcwd, the later uses a popen ("pwd"); */
+char *get_current_wd (char *buffer, int size)
+{
+    char *p;
+    
+#ifdef HAVE_GETWD
+    p = (char *) getwd (buffer);
+#else
+    p = getcwd (buffer, size);
+#endif
+    return p;
+}
+
+static char *get_absolute_name (char *file)
+{
+    char dir [MAXPATHLEN];
+
+    if (file [0] == '/')
+	return strdup (file);
+    get_current_wd (dir, MAXPATHLEN);
+    return get_full_name (dir, file);
+}
+
+int my_mkdir (char *s, mode_t mode)
+{
+    int result;
+
+    result = mkdir (s, mode);
+    if (result == 0){
+	s = get_absolute_name (s);
+	tree_add_entry (s);
+	free (s);
+    }
+    return result;
+}
+
+int my_rmdir (char *s)
+{
+    int result;
+
+    result = rmdir (s);
+    if (result == 0){
+	s = get_absolute_name (s);
+	tree_remove_entry (s);
+	free (s);
+    }
+    return result;
+}
